// Compile este arquivo com cekeikon5
// compila granul -pv -omp
#include <proeikon>
#include <imgpv>

struct ELEMENTO {
  double corr, raio; 
  int l, c; 
  char clas; // furo='f' quadrado='q'
  char aut; // auto='a' manual='m'
  //>> Dados temporarios
  int iraio;
  double infl; 
  double area;
};

struct BARRA {
  int num_radii;
  double init_radius;
  double final_radius; // acrescentado em 1.6
  double radii_per_octave;
  double passo; // acrescentado em 1.6
  int tcorr;
  int traio;
  int tinfl;
  int tquadrado;
  int tsobrep;
  char modorelat; // absoluto/relativo no relat
  vector<ELEMENTO> v;
  
  IMGGRY ent;
  pvCOR sai;
  string nomeimg;
  string nomehol;

  VETOR<double> sca;
  char modo; // insert, delete, furo/quadrado, auto/man no edit
  BARRA(); 
};

extern BARRA bar;

void savehol(string nome);
void loadhol(string nome);
void gerainfl();

BARRA::BARRA() { 
  tcorr=8; traio=10; tinfl=99; tquadrado=90; tsobrep=40;
  modorelat='r'; modo='f'; 
}

BARRA bar;

//<<<<<<<<<<<<<<< comum.cpp

void savehol(string nome)
{ int n=bar.v.size();
  double thresh=bar.tcorr/100.0; 
  int conta=0;
  for (int i=0; i<n; i++) 
    if (!(bar.v[i].corr<thresh)) conta++;

  FILE* arq=fopen(nome.c_str(),"wt");
  fprintf(arq,"num_radii=%d\n",bar.num_radii);
  fprintf(arq,"init_radius=%g\n",bar.init_radius);
  fprintf(arq,"radii_per_octave=%g\n",bar.radii_per_octave);
  fprintf(arq,"tcorr=%d\n",bar.tcorr);
  fprintf(arq,"traio=%d\n",bar.traio);
  fprintf(arq,"tinfl=%d\n",bar.tinfl);
  fprintf(arq,"tquadrado=%d\n",bar.tquadrado);
  fprintf(arq,"modorelat=%c\n",bar.modorelat);
  fprintf(arq,"n=%d\n",conta);

  for (int i=0; i<n; i++) 
    if (!(bar.v[i].corr<thresh)) {
      ELEMENTO& e=bar.v[i];
      fprintf(arq,"%g %g %d %d %c %c\n",e.corr,e.raio,e.l,e.c,e.clas,e.aut);
    }
  fclose(arq);
}

void loadhol(string nome)
{ FILE* arq=fopen(nome.c_str(),"rt");
  string st; int n; ELEMENTO e;

  st=LeStr(arq); if (st!="num_radii") erro("Erro: num_radii esperado"); LeNum(arq,bar.num_radii);
  st=LeStr(arq); if (st!="init_radius") erro("Erro: init_radius esperado"); LeNum(arq,bar.init_radius);
  st=LeStr(arq); if (st!="radii_per_octave") erro("Erro: radii_per_octave esperado"); LeNum(arq,bar.radii_per_octave);
  st=LeStr(arq); if (st!="tcorr") erro("Erro: tcorr esperado"); LeNum(arq,bar.tcorr);
  st=LeStr(arq); if (st!="traio") erro("Erro: traio esperado"); LeNum(arq,bar.traio);
  st=LeStr(arq); if (st!="tinfl") erro("Erro: tinfl esperado"); LeNum(arq,bar.tinfl);
  st=LeStr(arq); if (st!="tquadrado") erro("Erro: tquadrado esperado"); LeNum(arq,bar.tquadrado);
  st=LeStr(arq); if (st!="modorelat") erro("Erro: modorelat esperado"); 
    st=LeStr(arq); bar.modorelat=st.c_str()[0];
    if (bar.modorelat!='a' && bar.modorelat!='r') erro("Erro: modorelat deve ser a ou r");
  st=LeStr(arq); if (st!="n") erro("Erro: n esperado"); LeNum(arq,n);
  //debug(11);
  //printf("n=%d\n",n);

  for (int i=0; i<n; i++) { 
    LeNum(arq,e.corr); 
    LeNum(arq,e.raio); 
    LeNum(arq,e.l); 
    LeNum(arq,e.c); 
    st=LeStr(arq); e.clas=st[0];
    st=LeStr(arq); e.aut=st[0];
    bar.v.push_back(e); 
  }
  fclose(arq);
}

void gerainfl()
{ int n=bar.v.size();
  double mindist; double dist=infinito;
  for (int i=0; i<n; i++) {
    mindist=infinito;
    int l1=bar.v[i].l; int c1=bar.v[i].c; 
    for (int j=0; j<n; j++) {
      if (i!=j) {
        int l2=bar.v[j].l; int c2=bar.v[j].c; double r2=bar.v[j].raio;
        dist=sqrt(double(elev2(l1-l2)+elev2(c1-c2)))-r2;
        if (dist<mindist) mindist=dist;
      }
    }
    bar.v[i].infl=mindist;
  }
}

//<<<<<<<<<<<<<<<<<<<<<<<<<< corrcirc.cpp

namespace CorrCirc {

double f(double r, double scale)
{ if (r<=scale) return -1.0;
  else if (r<=sqrt(2.0)*scale) return 1.0;
  else return 0.0;
}

void cir(double scale, IMGFLT& sai)
{ for (int x=sai.minx(); x<=sai.maxx(); x++)
    for (int y=sai.miny(); y<=sai.maxy(); y++) {
      double r=sqrt(double(x*x+y*y));
      sai.atC(x,y)=f(r,scale);
    }
  sai=dcreject(sai);
  sai = (1.0/(M_2PI*elev2(scale)))*sai;
}

void quicksort(vector<double>& vetr, vector<int>& vets, vector<int>& vetl, vector<int>& vetc, int l, int r)
{ if (!(l<=r)) return; // caso vetor nulo
  int i=l; int j=r;
  double x=vetr[(l+r)/2];
  do {
    while (vetr[i]>x) i++;
    while (x>vetr[j]) j--;
    if (i<=j) { swap(vetr[i],vetr[j]); swap(vets[i],vets[j]); swap(vetl[i],vetl[j]); swap(vetc[i],vetc[j]); i++; j--; }
  } while (i<=j);
  if (l<j) quicksort(vetr,vets,vetl,vetc,l,j);
  if (i<r) quicksort(vetr,vets,vetl,vetc,i,r);
}

void empurra(vector<double>& vetr, vector<int>& vets, vector<int>& vetl, vector<int>& vetc, int i)
{ ELEMENTO e;
  e.corr=vetr[i];
  e.raio=bar.sca(vets[i]);
  e.l=vetl[i];
  e.c=vetc[i];
  e.clas='f'; // furo='f' quadrado='q'
  e.aut='a'; // auto='a' manual='m' 
  e.iraio=vets[i];
  e.infl=0.0; 
  bar.v.push_back(e);
}

bool existesobreposicao(vector<double>& vetr, vector<int>& vets, vector<int>& vetl, vector<int>& vetc, int i)
{ double s1=bar.sca(vets[i]); int l1=vetl[i]; int c1=vetc[i];
  for (int j=0; j<int(bar.v.size()); j++) {
    double s2=bar.v[j].raio; int l2=bar.v[j].l; int c2=bar.v[j].c;
    double dist=sqrt(double(elev2(l2-l1)+elev2(c2-c1)));
    double somaraio=s1+s2;
    if (dist-somaraio+(s1*bar.tsobrep/100.0)<0.0) return true;
  }
  return false;
}

void eliminasobreposto(vector<double>& vetr, vector<int>& vets, vector<int>& vetl, vector<int>& vetc)
// Elimina sobreposto e coloca em bar.v
{ bar.v.erase(bar.v.begin(),bar.v.end());
  int n=vetr.size();
  if (int(vets.size())!=n || int(vetl.size())!=n || int(vetc.size())!=n) erro("Erro inesperado 4");
  if (vetr.size()==0) return;
  quicksort(vetr,vets,vetl,vetc,0,n-1);
  empurra(vetr,vets,vetl,vetc,0);
  for (int i=1; i<n; i++) {
    if (!existesobreposicao(vetr,vets,vetl,vetc,i)) empurra(vetr,vets,vetl,vetc,i);
  }
}

IMGCOR gerasai()
{ IMGCOR sai=bar.ent; 
  int n=bar.v.size();
  double thresh=bar.tcorr/100.0; 
  for (int i=0; i<n; i++) {
    double corr=bar.v[i].corr;
    if (corr<thresh) continue;

    // desenha circulo
    int l=bar.v[i].l;
    int c=bar.v[i].c;
    int r=arredonda(bar.v[i].raio);
    //int s=bar.v[i].iraio;
    sai(l,c)=COR(255,0,0);
    circulo(sai,l,c,r,COR(255,0,0));
    //char st[256];
    //sprintf(st,"%d=%02d",s,arredonda(100*corr));
    //puttxt(sai,l,c-10,st,COR(0,0,255));
  }
  return sai;
}

void update(int arg)
{ IMGCOR sai=gerasai();
  IMGCOR2pvCOR(sai,bar.sai);
  cvShowImage( bar.nomeimg.c_str(), bar.sai.img );
}

vector<double> vetr;
vector<int> vets;
vector<int> vetl;
vector<int> vetc;
void updatesobreposto(int arg)
{ eliminasobreposto(vetr,vets,vetl,vetc);
  update(0);
}

void hardcopy(string nome)
{ IMGCOR sai=gerasai();
  pvImp(sai,nome);
}

}

void corrcirc_ori()
{ using namespace CorrCirc;

  int t1=centseg();
  printf(">>num_radii=%d init_radius=%f radii_per_octave=%f\n", bar.num_radii, bar.init_radius, bar.radii_per_octave);
  printf(">>passo=%f final_radius=%f\n", bar.passo, bar.final_radius);

  IMGGRY g; pvLe(g,bar.nomeimg);

  double passo=exp(log(2.0)/bar.radii_per_octave);
  I3DFLT f(bar.num_radii,g.nl(),g.nc());
  bar.sca.resize(bar.num_radii);
  { IMGFLT a=g;
    #pragma omp parallel for
    for (int s=0; s<bar.num_radii; s++) {
      IMGFLT mascara;
      bar.sca(s)=bar.init_radius*pow(passo,s);
      #pragma omp critical
      printf("s=%d sca(s)=%g\n",s,bar.sca(s));
      int n=teto(sqrt(2.0)*2*bar.sca(s)); if (n%2==0) n++;
      mascara.resize(n,n);
      cir(bar.sca(s),mascara);
      f(s)=pvConv(a,mascara,'s');
    }
    //imp(f,"f.img");
  }

  IMGFLT maxr(g.nl(),g.nc(),FLT(-infinito)); // maior correlacao
  IMGSHT maxs(g.nl(),g.nc(),SHT(0)); // escala que da maior correlacao
  #pragma omp parallel for
  for (int i=0; i<g.n(); i++) {
    for (int s=0; s<f.ns(); s++) {
      if (f(s)(i)>maxr(i)) { maxr(i)=f(s)(i); maxs(i)=s; }
    }
  }

  #pragma omp parallel for
  for (int l=1; l<g.nl()-1; l++) {
    for (int c=1; c<g.nc()-1; c++) {
      double valor=maxr(l,c);
      for (int l2=-1; l2<=1; l2++)
        for (int c2=-1; c2<=1; c2++) {
          if (valor<maxr(l+l2,c+c2)) 
            goto sai1; 
        }
      // E' maximo local
      #pragma omp critical 
      { vetr.push_back(maxr(l,c));
        vets.push_back(maxs(l,c));
        vetl.push_back(l);      
        vetc.push_back(c);
      }
      sai1:;
    }
  }
  printf("Numero de maximos locais=%u\n",vetr.size());
  int t=centseg()-t1;
  printf("Tempo gasto= %d.%02ds\n",t/100,t%100);

  bar.ent=g;
  cvNamedWindow(bar.nomeimg.c_str(),0);
  cvCreateTrackbar("TCorr", bar.nomeimg.c_str(), &bar.tcorr, 99, update); // 0 a 99
  cvCreateTrackbar("TSobrep", bar.nomeimg.c_str(), &bar.tsobrep, 99, updatesobreposto); // 0 a 99
  updatesobreposto(0);
  int k;
  do {
    printf("ESC=fim S=savehol H=hardcopy\n");
    k=tolower(cvWaitKey(0));
    if (k=='s') { printf("  SaveHol\n"); savehol(bar.nomehol); }
    else if (k=='h') { printf("  HardCopy corrcirc.png\n"); hardcopy("corrcirc.png"); }
  } while (k!=27);
}

void corrcirc(int argc, char** argv)
{ using namespace CorrCirc;
  if (argc!=6) {
    printf("CorrCirc: Extrema of circular kernel correlation to detect circular holes\n");
    printf("CorrCirc input.pgm output.hol num_radii init_radius radii_per_octave\n");
    erro("Error: Invalid number of arguments");
  }

  if (sscanf(argv[3],"%d",&bar.num_radii)!=1) erro("Error: Read bar.num_radii");
  if (bar.num_radii<1) erro("Error: bar.num_radii<1");

  if (sscanf(argv[4],"%lf",&bar.init_radius)!=1) erro("Error: Read bar.init_radius");
  if (bar.init_radius<0.3) erro("Error: bar.init_radius<0.3");

  if (sscanf(argv[5],"%lf",&bar.radii_per_octave)!=1) erro("Error: Read bar.radii_per_octave");
  if (bar.radii_per_octave<1) erro("Error: bar.radii_per_octave<1");

  bar.passo=exp(log(2.0)/bar.radii_per_octave);

  // double nr = r * log2(f/i) = r * (log(f/i) / log(2))   (eq1)
  // f = i * exp( n * log(2) / r )   (eq5)

  bar.final_radius = bar.init_radius * exp( bar.num_radii*log(2)/bar.radii_per_octave );  // isto nao e' necessario

  bar.nomeimg=string(argv[1]);
  bar.nomehol=string(argv[2]);
  corrcirc_ori();
}

void corrcirc2(int argc, char** argv)
{ using namespace CorrCirc;
  if (argc!=5 && argc!=6) {
    printf("CorrCirc: Extrema of circular kernel correlation to detect circular holes\n");
    printf("CorrCirc input.pgm output.hol init_radius final_radius [radii_per_octave=5]\n");
    printf("  Radii_per_octave may be modified to get integer number of radii.\n");
    erro("Error: Invalid number of arguments");
  }

  // Sejam double i=init_radius, double f=final_radius, double r=radii_per_octave, double nr=numero de raios real e int ni=num_radii
  // Sao dados i, f e r. Deve-se calcular nr real, arredondar para ni inteiro e modificar r para que ni de inteiro.
  // double nr = r * log2(f/i) = r * (log(f/i) / log(2))   (eq1)
  // Nota: Em C, log calcula log neperiano.
  // Exemplo: i=1, f=4, r=3. nr = 6.
  // int nd=arredonda(nr);  (eq2)
  // novo r = (ni * ln(2)) / ( ln(f/i) ) (eq3)
  // Exemplo: novo r = 3 
  // passo = exp( ln(f/i) / ni ) (eq4)

  if (sscanf(argv[3],"%lf",&bar.init_radius)!=1) erro("Error: init_radius");
  if (bar.init_radius<0.3) erro("Error: init_radius<0.3");

  if (sscanf(argv[4],"%lf",&bar.final_radius)!=1) erro("Error: final_radius");
  if (bar.final_radius<=bar.init_radius) erro("Error: final_radius<=init_radius");

  bar.radii_per_octave=5;
  if (argc>5) {
    if (sscanf(argv[5],"%lf",&bar.radii_per_octave)!=1) erro("Error: radii_per_octave");
    if (bar.radii_per_octave<1) erro("Error: bar.radii_per_octave<1");
  }

  double nr = bar.radii_per_octave * (log(bar.final_radius/bar.init_radius) / log(2)); // (eq1)
  bar.num_radii=arredonda(nr); // (eq2)
  bar.radii_per_octave = (bar.num_radii * log(2)) / ( log(bar.final_radius/bar.init_radius) ); // (eq3)
  bar.passo = exp( log(bar.final_radius/bar.init_radius) / bar.num_radii ); // (eq4)
  bar.num_radii++; //para que numero de raios conte o ultimo raio.

  bar.nomeimg=string(argv[1]);
  bar.nomehol=string(argv[2]);
  corrcirc_ori();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< classify.cpp

namespace Classify {

void classifica()
{ int n=bar.v.size();
  for (int i=0; i<n; i++) {
    //if (bar.v[i].raio>=bar.traio || bar.v[i].infl>=bar.tinfl)
    if (bar.v[i].raio>=bar.traio && bar.v[i].infl/bar.v[i].raio-1>=bar.tinfl/50.0)
      bar.v[i].clas='q';
    else 
      bar.v[i].clas='f';
  } 
}

IMGCOR gerasai()
{ IMGCOR sai=bar.ent; 
  int n=bar.v.size();
  for (int i=0; i<n; i++) {
    int l=bar.v[i].l;
    int c=bar.v[i].c;
    int s=arredonda(bar.v[i].raio);
    if (bar.v[i].aut=='m') {
      xis(sai,l,c,3,COR(0,255,0));
    }
    COR cor;
    if (bar.v[i].clas=='f') {
      cor=COR(255,0,0);
      sai(l,c)=cor; circulo(sai,l,c,s,cor);
    } else if (bar.v[i].clas=='q') {
      cor=COR(0,0,255);
      sai(l,c)=cor; circulo(sai,l,c,s,cor);
      circulo(sai,l,c,arredonda(bar.v[i].infl),COR(255,255,0));
    } else erro("Erro inesperado 3");
  }
  return sai;
}

void update(int arg)
{ classifica();
  IMGCOR sai=gerasai();
  IMGCOR2pvCOR(sai,bar.sai);
  cvShowImage( bar.nomeimg.c_str(), bar.sai.img );
}

void hardcopy(string nome)
{ classifica();
  IMGCOR sai=gerasai();
  pvImp(sai,nome);
}

}

void classify(int argc, char** argv)
{ using namespace Classify;
  if (argc!=3) {
    printf("Classify: Classifica automaticamente furos e quadrados\n");
    printf("Classify input.pgm inout.hol\n");
    erro("Error: Invalid number of arguments");
  }

  IMGGRY g; pvLe(g,argv[1]);

  loadhol(argv[2]);
  gerainfl();

  bar.ent=g;
  bar.nomeimg=argv[1];
  bar.nomehol=argv[2];
  cvNamedWindow(bar.nomeimg.c_str(),0);
  cvCreateTrackbar("TRaio", bar.nomeimg.c_str(), &bar.traio, 99, update); // 0 a 99
  cvCreateTrackbar("TInfl", bar.nomeimg.c_str(), &bar.tinfl, 99, update); // 0 a 99
  update(0);
  int k;
  do {
    printf("ESC=fim S=savehol H=hardcopy\n");
    k=tolower(cvWaitKey(0));
    if (k=='s') { printf("  SaveHol\n"); savehol(argv[2]); }
    else if (k=='h') { printf("  HardCopy classify.png\n"); hardcopy("classify.png"); }
  } while (k!=27);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< edit.cpp

namespace Edit {

IMGCOR gerasai()
{ IMGCOR sai=bar.ent; 
  int n=bar.v.size();
  for (int i=0; i<n; i++) {
    int l=bar.v[i].l;
    int c=bar.v[i].c;
    int s=arredonda(bar.v[i].raio);
    if (bar.v[i].aut=='m') {
      xis(sai,l,c,3,COR(0,255,0));
    }
    COR cor;
    if (bar.v[i].clas=='f') {
      cor=COR(255,0,0);
      sai(l,c)=cor; circulo(sai,l,c,s,cor);
    } else if (bar.v[i].clas=='q') {
      cor=COR(0,0,255);
      sai(l,c)=cor; circulo(sai,l,c,s,cor);
    } else erro("Erro inesperado 3");
  }
  return sai;
}

void update(int arg=0)
{ IMGCOR sai=gerasai();
  IMGCOR2pvCOR(sai,bar.sai);
  cvShowImage( bar.nomeimg.c_str(), bar.sai.img );
}

void hardcopy(string nome)
{ IMGCOR sai=gerasai();
  pvImp(sai,nome);
}

void on_mouse(int event, int c, int l, int flags, void* param)
{ static int l1=0, c1=0;
  if (bar.modo=='i') {
    if (event==CV_EVENT_LBUTTONDOWN) {
      l1=l; c1=c;
    } else if (event==CV_EVENT_LBUTTONUP) {
      ELEMENTO e;
      e.corr=1.0; e.raio=sqrt( double(elev2(l1-l)+elev2(c1-c)) );
      e.l=l1; e.c=c1; e.clas='q'; e.aut='m';
      bar.v.push_back(e);
      update();
    }
  } else if (bar.modo=='d') {
    if (event==CV_EVENT_LBUTTONDOWN) {
      l1=l; c1=c;
    } else if (event==CV_EVENT_LBUTTONUP) {
      // Converte circulo de diametro (l1,c1), (l,c) em circulo de centro (l1,c1) e raio r1
      double r1=sqrt(double(elev2(l-l1)+elev2(c-c1)))/2.0;
      l1=(l1+l)/2; c1=(c1+c)/2;
      // Apaga os quadrados e furos que tiverem interseccao com circulo de centro (l1,c1) e raio r1
      for (int i=0; i<int(bar.v.size()); i++) {
        int l2=bar.v[i].l; int c2=bar.v[i].c; double r2=bar.v[i].raio;
        double dist=sqrt(double(elev2(l1-l2)+elev2(c1-c2)));
        if (dist<=r1+r2) {
          vector<ELEMENTO>::iterator j=(vector<ELEMENTO>::iterator)&(bar.v[i]);
          bar.v.erase(j); 
        }
      }
      update();
    }
  } else if (bar.modo=='f') {
    if (event==CV_EVENT_LBUTTONDOWN) {
      for (int i=0; i<int(bar.v.size()); i++) {
        int l2=bar.v[i].l; int c2=bar.v[i].c;
        double r=bar.v[i].raio;
        double dist=sqrt(double(elev2(l-l2)+elev2(c-c2)));
        if (dist<=r) {
          if (bar.v[i].clas=='f') bar.v[i].clas='q'; 
          else if (bar.v[i].clas=='q') bar.v[i].clas='f'; 
          bar.v[i].aut='m';
          break;
        }
      }
      update();
    }
  } else if (bar.modo=='a') {
    if (event==CV_EVENT_LBUTTONDOWN) {
      for (int i=0; i<int(bar.v.size()); i++) {
        int l2=bar.v[i].l; int c2=bar.v[i].c;
        double r=bar.v[i].raio;
        double dist=sqrt(double(elev2(l-l2)+elev2(c-c2)));
        if (dist<=r) {
          if (bar.v[i].aut=='a') bar.v[i].aut='m'; 
          else if (bar.v[i].aut=='m') bar.v[i].aut='a'; 
          break;
        }
      }
      update();
    }
  } else {
    printf("modo=%c\n",bar.modo);
    erro("Erro inesperado: Modo invalido");
  }
}

}

void edit(int argc, char** argv)
{ using namespace Edit;
  if (argc!=3) {
    printf("Edit: Edita manualmente furos e quadrados\n");
    printf("Edit input.pgm inout.hol\n");
    erro("Error: Invalid number of arguments");
  }

  IMGGRY g; pvLe(g,argv[1]);

  loadhol(argv[2]);

  bar.ent=g;
  //bar.modo='f'; // Por que precisa deste comando, se ha o mesmo no construtor?
  bar.nomeimg=argv[1];
  bar.nomehol=argv[2];
  cvNamedWindow(bar.nomeimg.c_str(),0);
  cvSetMouseCallback(bar.nomeimg.c_str(),on_mouse);
  update();
  int k;
  do {
    printf("ESC=fim S=savehol H=hardcopy I=insert D=delete F=furo/quad A=auto/man. Modo=[%c]\n",toupper(bar.modo));
    k=tolower(cvWaitKey(0));
    if (k=='s') { printf("  SaveHol\n"); savehol(argv[2]); }
    else if (k=='h') { printf("  HardCopy edit.png\n"); hardcopy("edit.png"); }
    else if (k=='i') { printf("  Modo insert\n"); bar.modo='i'; }
    else if (k=='d') { printf("  Modo delete\n"); bar.modo='d'; }
    else if (k=='f') { printf("  Modo furo/quadrado\n"); bar.modo='f'; }
    else if (k=='a') { printf("  Modo auto/manual\n"); bar.modo='a'; }
  } while (k!=27);
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< relat.cpp

namespace Relat {

int achanivelapropriado(int k)
{ return bar.tquadrado; 
}

void cresceabs(IMGGRY& quadrado, int k, double& area)
// bar.ent, bar.v[k].l, bar.v[k].c, bar.v[k].raio sao passados implicitamente
{ quadrado.backg()=255; bar.ent.backg()=255;
  int nivelcinza=achanivelapropriado(k);
  queue<int> q; q.push(bar.v[k].l); q.push(bar.v[k].c);
  int tamquadrado=0;
  while (!q.empty()) {
    int l=q.front(); q.pop();
    int c=q.front(); q.pop();
    double distancia=sqrt(double(elev2(l-bar.v[k].l)+elev2(c-bar.v[k].c)));
    if (quadrado(l,c)==0 && bar.ent(l,c)<nivelcinza && distancia<1.3*bar.v[k].raio) { 
      quadrado(l,c)=255; tamquadrado++;
      q.push(l-1); q.push(c);
      q.push(l+1); q.push(c);
      q.push(l);   q.push(c-1);
      q.push(l);   q.push(c+1);
    }
  }
  area=tamquadrado;
}

void crescerel(IMGGRY& quadrado, int k, double& area)
// bar.ent, bar.v[k].l, bar.v[k].c, bar.v[k].raio sao passados implicitamente
{ quadrado.backg()=255; bar.ent.backg()=255;
  int nivelcinza=achanivelapropriado(k);
  queue<int> q; q.push(bar.v[k].l); q.push(bar.v[k].c);
  int soma=0; int tamquadrado=0;
  while (!q.empty()) {
    int l=q.front(); q.pop();
    int c=q.front(); q.pop();
    double distancia=sqrt(double(elev2(l-bar.v[k].l)+elev2(c-bar.v[k].c)));
    if (tamquadrado==0 || (quadrado(l,c)==0 && (bar.ent(l,c)-arredonda(soma,tamquadrado))<nivelcinza && distancia<1.3*bar.v[k].raio)) { 
      quadrado(l,c)=255;
      soma+=bar.ent(l,c); tamquadrado++;
      q.push(l-1); q.push(c);
      q.push(l+1); q.push(c);
      q.push(l);   q.push(c-1);
      q.push(l);   q.push(c+1);
    }
  }
  area=tamquadrado;
}

IMGCOR gerasai()
{ IMGGRY quadrado(bar.ent.nl(),bar.ent.nc(),0);

  IMGCOR sai=bar.ent; 
  int n=bar.v.size();
  int k;
  for (k=0; k<n; k++) {
    if (bar.v[k].clas=='q') {
      if      (bar.modorelat=='r') crescerel(quadrado,k,bar.v[k].area); // bar.ent, bar.v[k] sao passados implicitamente
      else if (bar.modorelat=='a') cresceabs(quadrado,k,bar.v[k].area); // bar.ent, bar.v[k] sao passados implicitamente
    }
  }

  for (int i=0; i<sai.n(); i++)
    if (quadrado(i)==255) { BYTE g=sai(i).b(); sai(i)=COR(255,g,g); }

  for (int i=0; i<n; i++) {
    if (bar.v[i].clas=='q') {
      int l=bar.v[i].l;
      int c=bar.v[i].c;
      int s=arredonda(bar.v[i].raio);
      COR cor=COR(0,0,255);
      sai(l,c)=cor; circulo(sai,l,c,s,cor);
      if (bar.v[i].aut=='m') {
        xis(sai,l,c,3,COR(0,255,0));
      }
    }
  }

  return sai;
}


void update(int arg)
{ IMGCOR sai=gerasai();
  IMGCOR2pvCOR(sai,bar.sai);
  cvShowImage( bar.nomeimg.c_str(), bar.sai.img );
}

void hardcopy(string nome)
{ IMGCOR sai=gerasai();
  pvImp(sai,nome);
}

/*
bool menor1(ELEMENTO& a, ELEMENTO& b)
// Ordena chave-primaria=furo/quadrado (v[k].clas) e chave-secundaria=raio (v[k].area)
{ if (a.clas<b.clas) return true;
  if (a.clas==b.clas && a.area<b.area) return true;
  return false;
}

void ordena1(int l, int r)
// Ordena chave-primaria=furo/quadrado (v[k].clas) e chave-secundaria=area (v[k].area)
{ if (!(l<=r)) return; // caso vetor nulo
  int i=l; int j=r;
  ELEMENTO x=bar.v[(l+r)/2];
  do {
    while (menor1(bar.v[i],x)) i++;
    while (menor1(x,bar.v[j])) j--;
    if (i<=j) { swap(bar.v[i],bar.v[j]); i++; j--; }
  } while (i<=j);
  if (l<j) ordena1(l,j);
  if (i<r) ordena1(i,r);
}
*/

void calculaareafuro()
// calcula area quando bar.v.clas=='f'
{ int n=bar.v.size();
  for (int i=0; i<n; i++)
    if (bar.v[i].clas=='f')
      bar.v[i].area=M_PI*elev2(bar.v[i].raio);
}

void printrel(string nome)
{ gerainfl();
  calculaareafuro();
  int n=bar.v.size();
  //ordena1(0,n-1);

  // Gera histograma dos furos e quadrados
  VETOR<int> histfuro(bar.num_radii,0);
  VETOR<int> histquad(bar.num_radii,0);
  for (int i=0; i<n; i++) {
    double r=sqrt(bar.v[i].area/M_PI);
    int j=arredonda(bar.radii_per_octave*(log2(r)-log2(bar.init_radius)));
    if (j<0) j=0; if (j>=bar.num_radii) j=bar.num_radii-1;
    if      (bar.v[i].clas=='f') histfuro(j)++;
    else if (bar.v[i].clas=='q') histquad(j)++;
    else erro("Erro: classe invalido");
  }

  // Gera histograma das zonas de influencias
  int maxj=0;
  for (int i=0; i<n; i++) {
    if (bar.v[i].clas=='q') {
      int j=arredonda(bar.radii_per_octave*(log2(bar.v[i].infl)-log2(bar.init_radius)));
      if (j>maxj) maxj=j;
    }
  }
  VETOR<int> histinfl(maxj+1,0);
  for (int i=0; i<n; i++) {
    if (bar.v[i].clas=='q') {
      int j=arredonda(bar.radii_per_octave*(log2(bar.v[i].infl)-log2(bar.init_radius)));
      if (j<0) j=0;
      histinfl(j)++;
    }
  }

  double areafuro=0.0;
  double areaquad=0.0;
  for (int i=0; i<n; i++) {
    if      (bar.v[i].clas=='f') areafuro+=bar.v[i].area;
    else if (bar.v[i].clas=='q') areaquad+=bar.v[i].area;
  }

  double passo=exp(log(2.0)/bar.radii_per_octave);
  FILE* arq=fopen(nome.c_str(),"wt");

  //for (int i=0; i<n; i++) {
  //  fprintf(arq,"clas=%c area=%g raio=%g infl=%g\n",bar.v[i].clas,bar.v[i].area,bar.v[i].raio,bar.v[i].infl);
    //if (bar.v[i].clas=='f') fprintf(arq,"%g \n",bar.v[i].area);
  //}
  //fprintf(arq,"\n");
  fprintf(arq,"Area total da figura=%d pixels.\n",bar.ent.n());
  fprintf(arq,"Area total dos furos circulares=%g pixels.\n",areafuro);
  fprintf(arq,"Area total dos furos quadrados=%g pixels.\n",areaquad);
  fprintf(arq,"\n");

  fprintf(arq,"Histograma dos furos circulares\n");
  fprintf(arq,"%8s %8s %8s %8s\n","indice","raio","area","qtidade");
  for (int i=0; i<bar.num_radii; i++) {
    double raio=bar.init_radius*pow(passo,i);
    double area=M_PI*elev2(raio);
    fprintf(arq,"%8d %8.2f %8.2f %8d\n",i,raio,area,histfuro(i));
  }
  fprintf(arq,"\n");

  fprintf(arq,"Histograma dos furos quadrados\n");
  fprintf(arq,"%8s %8s %8s %8s\n","indice","raio","area","qtidade");
  for (int i=0; i<bar.num_radii; i++) {
    double raio=bar.init_radius*pow(passo,i);
    double area=M_PI*elev2(raio);
    fprintf(arq,"%8d %8.2f %8.2f %8d\n",i,raio,area,histquad(i));
  }
  fprintf(arq,"\n");

  fprintf(arq,"Histograma das zonas de infl. dos f. quad.\n");
  fprintf(arq,"%8s %8s %8s %8s\n","indice","raio","area","atidade");
  for (int i=0; i<histinfl.n(); i++) {
    double raio=bar.init_radius*pow(passo,i);
    double area=M_PI*elev2(raio);
    fprintf(arq,"%8d %8.2f %8.2f %8d\n",i,raio,area,histinfl(i));
  }

  fclose(arq);
}

} // namespace

void relat(int argc, char** argv)
{ using namespace Relat;
  if (argc!=4) {
    printf("Relat: Gera relatorio\n");
    printf("Relat input.pgm input.hol relat.txt\n");
    erro("Error: Invalid number of arguments");
  }

  IMGGRY g; pvLe(g,argv[1]);
  loadhol(argv[2]);

  bar.ent=g;
  bar.nomeimg=string(argv[1]);
  bar.nomehol=string(argv[2]);
  cvNamedWindow(bar.nomeimg.c_str(),0);
  cvCreateTrackbar("TQuadrado", bar.nomeimg.c_str(), &bar.tquadrado, 255, update); // 0 a 255
  update(0);
  int k;
  do {
    printf("ESC=fim S=savehol P=printrel A=cres abs R=cres rel. Modorelat=%c\n",toupper(bar.modorelat));
    k=tolower(cvWaitKey(0));
    if (k=='s')      { printf("  SaveHol\n"); savehol(argv[2]); }
    else if (k=='h') { printf("  HardCopy relat.png\n"); hardcopy("relat.png"); }
    else if (k=='p') { printf("  PrintRel\n"); printrel(argv[3]); }
    else if (k=='r') { printf("  Modo limiarizacao relativo\n"); bar.modorelat='r'; update(0); }
    else if (k=='a') { printf("  Modo limiarizacao absoluto\n"); bar.modorelat='a'; update(0); }
  } while (k!=27);
}

//<<<<<<<<<<<<<<<<<<<<<<<< main.cpp

void corrcirc(int argc, char** argv);
void classify(int argc, char** argv);
void edit(int argc, char** argv);
void relat(int argc, char** argv);

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< main <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
const char about[]=
"< Granul.exe: Granulometria de correlacao para silicio poroso, versao 1.6>\n"
"_______________________________________________________________________________\n";

const char help[]=
"Programas:\n"
"  CorrCirc  - Extremas das correlacoes com mascaras circulares\n"
"  CorrCirc2 - Extremas das correlacoes - raio inicial e final como parametros\n"
"  Classify  - Classifica automaticamente furos e quadrados\n"
"  Edit      - Edita manualmente furos/quadrados\n"
"  Relat     - Gera relatorio\n"
"...............................................................................\n";

int main(int argc, char** argv)
{ bar=BARRA();
  //debug(-2);
  fprintf(stderr,about);
  if (argc<2) {
    cout << help;
    erro("Erro: Numero de argumentos invalido");
  } else {
    string comando=strlwr(argv[1]);
    if      (comando=="corrcirc") corrcirc(argc-1,&argv[1]);
    else if (comando=="corrcirc2") corrcirc2(argc-1,&argv[1]);
    else if (comando=="classify") classify(argc-1,&argv[1]);
    else if (comando=="edit")     edit(argc-1,&argv[1]);
    else if (comando=="relat")    relat(argc-1,&argv[1]);
    else erro("Erro: Programa inexistente ",comando.c_str());
  }
}




